@@ -1,11 +1,14 @@
 #include "ACProject.h"
 #include "ACConfig.h"
 #include "version.h"
-#include "Puma/SysCall.h"
-using namespace Puma;
+#include "SysCall.h"
 
 #ifdef FRONTEND_CLANG
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+#include "llvm/TargetParser/Host.h"
+#else
 #include "llvm/Support/Host.h"
+#endif
 #include "clang/Basic/Version.h"
 #include "clang/Basic/TargetInfo.h"
 #include "clang/Frontend/CompilerInstance.h"
@@ -21,21 +24,47 @@ using namespace clang;
 #include <libxml/xpath.h>
 #include <stdio.h>
 #include <fstream>
+#include <regex>
 
 #ifdef FRONTEND_CLANG
 ACProject::ACProject (ACErrorStream &err, int &argc, char **&argv)
-    : PathManager (err), _ci (0), _puma_config (err) {
+    : _err (err), _ci (0), _puma_config (_puma_err), _tricore_hack(false) {
 
   if (!_puma_config.CustomSystemConfigFile (argc, argv))
     _puma_config.Read (); // read global config file, e.g. $PUMA_CONFIG
 
   _puma_config.Read (argc, argv); // read command line config arguments
 
+  // determine the supported C++ versions
+  _backend_standard = -1; // unknown standard
+  for (unsigned i = 0; i < _puma_config.Options (); i++) {
+    const Puma::ConfOption *o = _puma_config.Option (i);
+    if (!strcmp(o->Name(), "-D")) {
+      if (strcmp (o->Argument(0), "__cplusplus") == 0) {
+        if (strcmp (o->Argument(1), "199711L") == 0)
+          _backend_standard = 1998;
+        else if (strcmp (o->Argument(1), "201103L") == 0) // is it C++11?
+          _backend_standard = 2011;
+        else if (strcmp (o->Argument(1), "201300L") == 0 ||
+            strcmp (o->Argument(1), "201402L") == 0)  // is it C++14?
+          _backend_standard = 2014;
+        else if (strcmp (o->Argument(1), "201406L") == 0 ||
+            strcmp (o->Argument(1), "201500L") == 0 ||
+            strcmp (o->Argument(1), "201703L") == 0) // is it C++17?
+          _backend_standard = 2017;
+        else if (strcmp (o->Argument(1), "201707L") == 0 ||
+            strcmp (o->Argument(1), "201709L") == 0 ||
+            strcmp (o->Argument(1), "202002L") == 0) // is it C++20?
+          _backend_standard = 2020;
+      }
+    }
+  }
+
   // read command line arguments
   config ().Add ("-D", "__acweaving", ""); // only defined during weaving phase
 
   // configure path manager
-  configure (_puma_config);
+  configure_project_file_manager (_puma_config);
 }
 #else
 ACProject::ACProject (ACErrorStream &err, int &argc, char **&argv)
@@ -46,10 +75,107 @@ ACProject::ACProject (ACErrorStream &err, int &argc, char **&argv)
 #endif
 
 #ifdef FRONTEND_CLANG
+
+class PatchedTargetInfo : public clang::TargetInfo {
+  
+private:
+  void replaceOrAppendDL(std::string &dl, std::string entry, std::string newEntry) {
+      // find entry
+      std::size_t pos = dl.find(entry);
+      
+      if (pos == std::string::npos) {
+        // No entry present; append one.
+        if (!dl.empty() && dl.back() != '-')
+          dl.push_back('-');
+
+        dl += newEntry;
+        return;
+      }
+
+      // Replace the existing entry:: segment until the next '-' or end.
+      std::size_t end = dl.find('-', pos);
+      if (end == std::string::npos)
+        dl.replace(pos, dl.size() - pos, newEntry);
+      else
+        dl.replace(pos, end - pos, newEntry);
+    };
+
+public:
+  void patchTC() {
+    // get DL string
+    std::string DL = std::string(getDataLayoutString());
+
+    // patch size type
+    SizeType = clang::TransferrableTargetInfo::IntType::UnsignedLong;
+
+    // patch long double
+    LongDoubleWidth = 8 * 8;
+    //LongDoubleAlign = 4 * 8;
+    //LongDoubleFormat = &llvm::APFloat::IEEEdouble();
+
+    // patch double
+    DoubleAlign = 4 * 8;
+    replaceOrAppendDL(DL, "f64:", "f64:32:32");
+
+    // patch long long
+    LongLongAlign = 4 * 8;
+    replaceOrAppendDL(DL, "i64:", "i64:32:32");
+
+    // reset DL 
+    resetDataLayout(DL);
+  }
+};
+
 ACProject::~ACProject() {
   delete _ci;
 }
-#endif
+
+// Configure the project from the command line or a file.
+void ACProject::configure_project_file_manager(const Puma::Config &c) {
+  const Puma::ConfOption *d = 0, *p = 0;
+
+  unsigned num = c.Options();
+  for (unsigned i = 0; i < num; i++) {
+    const Puma::ConfOption *o = c.Option(i);
+    bool new_p = false, new_d = false;
+
+    if (!strcmp(o->Name(), "-w")) {
+      if (o->Arguments() != 1)
+        continue;
+      protect(o->Argument(0));
+    } else if (!strcmp(o->Name(), "-p")) {
+      if (o->Arguments() != 1)
+        continue;
+      new_p = true;
+    } else if (!strcmp(o->Name(), "-d")) {
+      if (o->Arguments() != 1)
+        continue;
+      new_d = true;
+    }
+
+    if (new_p) {
+      if (p) {
+        addPath(p->Argument(0), d ? d->Argument(0) : 0);
+        if (d)
+          d = 0;
+      }
+      p = o;
+    }
+
+    if (new_d) {
+      if (d) {
+        addPath(p ? p->Argument(0) : 0, d->Argument(0));
+        if (p)
+          p = 0;
+      }
+      d = o;
+    }
+  }
+
+  if (p || d)
+    addPath(p ? p->Argument(0) : 0, d ? d->Argument(0) : 0);
+}
+#endif // FRONTEND_CLANG
 
 void ACProject::addFiles (xmlDocPtr doc, xmlNodePtr node, const string& xpath) {
   xmlXPathContextPtr  m_Context = xmlXPathNewContext(doc);
@@ -73,15 +199,18 @@ void ACProject::addFiles (xmlDocPtr doc, xmlNodePtr node, const string& xpath) {
             if (last != '/' && last != '\\')
               destname += '/';
 #ifdef WIN32
-            SysCall::MakeUnixPath ((char*)relpath);
-#endif
+            std::string relpath_unix = (const char*)relpath;
+            SysCall::make_unix_path(relpath_unix);
+            destname += relpath_unix;
+#else
             destname += (const char*)relpath;
+#endif
           }
           else
             destname = "<unused-dest-name>";
           
           // add the file to the project
-          PathManager::addFile ((const char*)path, destname.c_str ());
+          addFile ((const char*)path, destname.c_str ());
         }
         else {
           assert (false);
@@ -111,11 +240,15 @@ ACFileID ACProject::addVirtualFile (const string &filename, const string &conten
   // FIXME: Hack
 //  fclose (fopen (filename.c_str (), "w+"));
   clang::FileManager &fm = _ci->getFileManager ();
-  const clang::FileEntry *fe = fm.getVirtualFile (filename, 0, 0);
-  _ci->getSourceManager().overrideFileContents(fe,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+  ACFileID fid (fm.getVirtualFileRef (filename, 0, 0));
+#else
+  ACFileID fid (fm.getVirtualFile (filename, 0, 0));
+#endif
+  _ci->getSourceManager().overrideFileContents(fid.file_entry (),
       llvm::MemoryBuffer::getMemBufferCopy(contents));
-  _virtual_files.insert(ACFileID(fe));
-  return fe;
+  _virtual_files.insert(fid);
+  return fid;
 #endif
 }
 
@@ -156,7 +289,11 @@ class Diag : public TextDiagnosticPrinter {
   llvm::raw_string_ostream _str_stream;
   llvm::raw_ostream &_orig_stream;
 public:
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_21_0_0
+  Diag(raw_ostream &os, DiagnosticOptions &diags) :
+#else
   Diag(raw_ostream &os, DiagnosticOptions *diags) :
+#endif
     TextDiagnosticPrinter(_str_stream, diags), _str_stream(_msg),
     _orig_stream (os) {}
   void HandleDiagnostic (DiagnosticsEngine::Level DiagLevel,
@@ -164,6 +301,8 @@ public:
     // call Clang's HandleDiagnostic function. Output will be written into _msg.
     TextDiagnosticPrinter::HandleDiagnostic(DiagLevel, Info);
 
+    // cout << "Handle Diag: " << endl << _msg << endl << "Handle Diag End" << endl;
+    _msg = std::regex_replace(_msg, std::regex("(^|\r?\n)(.*<built-in>:.*)(\r?\n)"), "");
     // search for <intro:address> in _msg and replace it with a proper filename
     for (unsigned pos = 0; pos < _msg.size(); pos++) {
       if (_msg.substr(pos, 7) == "<intro:") {
@@ -237,18 +376,34 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
   Args.push_back("-fexceptions");
   Args.push_back("-pic-level");
   Args.push_back("2");
+#if CLANG_VERSION_NUMBER < VERSION_NUMBER_18_1_3
   Args.push_back("-ftemplate-depth");
   Args.push_back("65536");
+#endif
   Args.push_back("-fblocks");
 
+  // disable all diagnostics: this is up the backend compiler
+  Args.push_back("-Wno-everything");
+
+  // set arg for configuring the C++ standard
+  switch (_backend_standard) {
+    case 1998: Args.push_back("-std=c++98"); break;
+    case 2011: Args.push_back("-std=c++11"); break;
+    case 2014: Args.push_back("-std=c++14"); break;
+    case 2017: Args.push_back("-std=c++17"); break;
+    case 2020: Args.push_back("-std=c++20"); break;
+  }
+
   // Translate Puma's built-in config file into clang arguments.
   // Puma already sorts the options and gives system options a lower
   // priority than user-provided options.
-  Config &c = _puma_config;
+  Puma::Config &c = _puma_config;
+
+  bool need_neon_support = false;
 
   // Start with all system options that are not include paths
   for (unsigned i = 0; i < c.Options (); i++) {
-    const ConfOption *o = c.Option (i);
+    const Puma::ConfOption *o = c.Option (i);
     const char *Name = o->Name();
 
     // For things like --isystem.
@@ -256,35 +411,54 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
       Name += 1; // Skip first '-'.
 
     if (!strcmp(Name, "-D")) {
-      // Enable C++11/14 support if __cplusplus is defined and has the right value
-      if (strcmp (o->Argument(0), "__cplusplus") == 0) {
-        if (strcmp (o->Argument(1), "199711L") == 0)
-          Args.push_back("-std=c++98");
-        else if (strcmp (o->Argument(1), "201103L") == 0) // is it C++11?
-          Args.push_back("-std=c++11");
-        else if (strcmp (o->Argument(1), "201300L") == 0 ||
-            strcmp (o->Argument(1), "201402L") == 0)  // is it C++14?
-          Args.push_back("-std=c++14");
-        else if (strcmp (o->Argument(1), "201406L") == 0 ||
-            strcmp (o->Argument(1), "201500L") == 0 ||
-            strcmp (o->Argument(1), "201703L") == 0) // is it C++17?
-          Args.push_back("-std=c++17");
-        else if (strcmp (o->Argument(1), "201707L") == 0 ||
-            strcmp (o->Argument(1), "201709L") == 0 ||
-            strcmp (o->Argument(1), "202002L") == 0) // is it C++20?
-          Args.push_back("-std=c++20");
-        continue; // no need to copy this macro, clang defines it anyway
-      }
       // Disable definition of macros that Clang will define anyway -- even with "UsePredefines = false".
-      if (strcmp (o->Argument(0), "__STDC__") == 0 ||
+      if (strcmp (o->Argument(0), "__cplusplus") == 0 ||
+          strcmp (o->Argument(0), "__STDC__") == 0 ||
           strcmp (o->Argument(0), "__STDC_HOSTED__") == 0 ||
           strcmp (o->Argument(0), "__STDC_VERSION__") == 0 ||
           strcmp (o->Argument(0), "__STDC_UTF_16__") == 0 ||
           strcmp (o->Argument(0), "__STDC_UTF_32__") == 0 ||
+          strcmp (o->Argument(0), "__STDC_EMBED_EMPTY__") == 0 ||
+          strcmp (o->Argument(0), "__STDC_EMBED_FOUND__") == 0 ||
+          strcmp (o->Argument(0), "__STDC_EMBED_NOT_FOUND__") == 0 ||
           strcmp (o->Argument(0), "__STDCPP_DEFAULT_NEW_ALIGNMENT__") == 0 ||
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+          // TODO: it would be better to check whether *any* macros from the config file are always defined by clang
+          strcmp (o->Argument(0), "__FLT_EVAL_METHOD__") == 0 ||
+          strcmp (o->Argument(0), "__STDCPP_THREADS__") == 0 ||
+#endif
           strcmp (o->Argument(0), "__has_include(STR)") == 0 ||
           strcmp (o->Argument(0), "__has_include_next(STR)") == 0)
         continue;
+
+      // hacks to work around clang bugs/incompatibilities ...
+      if (strcmp (o->Argument(0), "__GNUC__") == 0) {
+        // "Missing support for GCC's attribute malloc (deallocator) and malloc (deallocator, ptr-index)" #53152
+        // if __GNUC__ is >= 11 we remove the arguments of __malloc__ by a macro
+        if (o->Arguments() == 2 && atoi(o->Argument(1)) >= 11) {
+          Args.push_back("-D __malloc__(A,B)=__malloc__");
+        }
+        // g++-13 support _Float{32|64|128} natively, but clang does not
+        if (o->Arguments() == 2 && atoi(o->Argument(1)) >= 13) {
+          Args.push_back("-D __acdef_float");
+        }
+      }
+
+      // clang doesn't support the new float suffixes (e.g. in "typedef __decltype(0.0bf16) __bfloat16_t;"), yet
+      if (strncmp(o->Argument(0), "__BFLT16_", 9) == 0 ||
+          strncmp(o->Argument(0), "__FLT32_", 8) == 0 ||
+          strncmp(o->Argument(0), "__FLT64_", 8) == 0 ||
+          strncmp(o->Argument(0), "__FLT128_", 9) == 0) {
+        continue;
+      }
+
+      // enable Arm NEON if needed
+      if (strncmp(o->Argument(0), "__ARM_NEON", 10) == 0) {
+        if (o->Arguments() == 2 && strncmp(o->Argument(1), "1", 1) == 0) {
+          need_neon_support = true;
+        }
+      }
+
       // Don't define the __puma macro when Clang parses the code
       if (strcmp (o->Argument(0), "__puma") == 0)
         continue;
@@ -296,11 +470,24 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
     // replace the --target option by -triple
     else if (!strcmp(Name,"--target")) {
       Args.push_back("-triple");
-      Args.push_back(o->Argument(0));
+      // clang has no tricore target
+      if(strncmp(o->Argument(0), "tricore", 7) == 0){
+        // enable tricore hack
+        _tricore_hack = true;
+
+        // use elfiamcu as base because its similar to tricore and has allowsLargerPreferedTypeAlignment as false
+        Args.push_back("i386-pc-elfiamcu");
+      }
+      else
+        Args.push_back(o->Argument(0));
     }
     // if gnu extensions shall be supported, enable the necessary clang parser options
     else if (!strcmp(Name,"--gnu")) {
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_20_0_0
+      Args.push_back("-Wno-error=invalid-gnu-asm-cast"); // see bug 766 (ignore strange casts of LValues)
+#else
       Args.push_back("-fheinous-gnu-extensions"); // see bug 766 (ignore strange casts of LValues)
+#endif
     }
 
     // Pass macro definitions, include paths, and framework paths to clang. Everything else is
@@ -308,32 +495,78 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
     if (strncmp(Name, "-I", 2) && strncmp(Name, "-i", 2) && strncmp(Name, "-F", 2))
       continue;
 
-    Args.push_back(Name);
-    for (unsigned j = 0; j < o->Arguments (); j++)
-      Args.push_back(o->Argument (j));
+    bool ignore_path = false;
+    for (unsigned i2 = i+1; i2 < c.Options(); i2++) {
+      const Puma::ConfOption *o2 = c.Option (i2);
+      if (strcmp (o2->Name(), "--isystem") == 0 && strcmp(o->Argument(0), o2->Argument(0)) == 0) {
+        ignore_path = true;
+        break;
+      }
+    }
+
+    if (!ignore_path) {
+      Args.push_back(Name);
+      // cout << "Arg: " << Name;
+      for (unsigned j = 0; j < o->Arguments (); j++) {
+        Args.push_back(o->Argument (j));
+        // cout << " " << o->Argument (j);
+      }
+      // cout << endl;
+    }
+    // else
+    //   cout << "system path ignored" << endl;
+  }
+
+  // if g++ has __ARM_NEON=1 defined, we must tell clang to support SIMD types
+  if (need_neon_support) {
+    Args.push_back("-target-feature");
+    Args.push_back("+simd");
   }
 
   // Add the main TU.
   Args.push_back("-x");
   Args.push_back("c++");
-  Args.push_back(conf.file_in());
+  Args.push_back(conf.file_in().c_str());
 
-  // Now create a -cc1 invocation out of the config.
-  CompilerInstance *new_ci = new CompilerInstance;
   // Create diagnostic (= diagnostics engine) (which is used for emitting errors, warnings
   // and so on): We use the adapted diagnostic "Diag" (the class "Diag" is defined above)
   // which replaces file names and line numbers of code that was injected by AspectC++
   // before outputting it to the console / before showing it to the user.
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_21_0_0
+  auto fs = llvm::vfs::getRealFileSystem();
+  // Now create a -cc1 invocation out of the config.
+  clang::DiagnosticOptions diag_opts;
+  auto diags = CompilerInstance::createDiagnostics(*fs, diag_opts, new Diag(llvm::errs(), diag_opts));
+  shared_ptr<CompilerInvocation> CI(new CompilerInvocation());
+  CompilerInvocation::CreateFromArgs(*CI, Args, *diags);
+  CompilerInstance *new_ci = new CompilerInstance(CI);
+#elif CLANG_VERSION_NUMBER >= VERSION_NUMBER_20_0_0
+  auto fs = llvm::vfs::getRealFileSystem();
+  // Now create a -cc1 invocation out of the config.
+  CompilerInstance *new_ci = new CompilerInstance;
+  auto diags = CompilerInstance::createDiagnostics(*fs, &new_ci->getDiagnosticOpts(), new Diag(llvm::errs(), &new_ci->getDiagnosticOpts()));
+  shared_ptr<CompilerInvocation> CI(new CompilerInvocation());
+  CompilerInvocation::CreateFromArgs(*CI, Args, *diags);
+  new_ci->setInvocation(std::move(CI));
+#else
+  // Now create a -cc1 invocation out of the config.
+  CompilerInstance *new_ci = new CompilerInstance;
   new_ci->createDiagnostics(new Diag(llvm::errs(), &new_ci->getDiagnosticOpts()));
   shared_ptr<CompilerInvocation> CI(new CompilerInvocation());
   CompilerInvocation::CreateFromArgs(*CI, Args, new_ci->getDiagnostics());
   new_ci->setInvocation(std::move(CI));
+#endif
 
   // Create the diagnostics engine again (see above)
   // Note: it is important to do this *after* calling 'setInvocation', because otherwise
   //       the diagnostics options are uninitialized / all false
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_21_0_0
+  new_ci->createDiagnostics(*fs, new Diag(llvm::errs(), new_ci->getDiagnosticOpts()));
+#elif CLANG_VERSION_NUMBER >= VERSION_NUMBER_20_0_0
+  new_ci->createDiagnostics(*fs, new Diag(llvm::errs(), &new_ci->getDiagnosticOpts()));
+#else
   new_ci->createDiagnostics(new Diag(llvm::errs(), &new_ci->getDiagnosticOpts()));
-
+#endif
   // Create vital components.
   if (_ci) {
     new_ci->setFileManager (&_ci->getFileManager ()); // reference in _ci must be set to 0 later!
@@ -346,9 +579,20 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
   new_ci->createSourceManager(new_ci->getFileManager());
 //  new_ci->setTarget(clang::TargetInfo::CreateTargetInfo(new_ci->getDiagnostics(),
 //                                                        &new_ci->getTargetOpts()));
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_21_0_0
+  clang::TargetOptions &targetOptions = new_ci->getTargetOpts();
+#else
   const std::shared_ptr<clang::TargetOptions> targetOptions = std::make_shared<clang::TargetOptions>(new_ci->getTargetOpts());
+#endif
 //  targetOptions->Triple = llvm::sys::getDefaultTargetTriple();
   clang::TargetInfo *pTargetInfo = clang::TargetInfo::CreateTargetInfo(new_ci->getDiagnostics(), targetOptions);
+
+  // apply tricore patch if necessary
+  if(_tricore_hack) {
+    PatchedTargetInfo *pti = (PatchedTargetInfo*)pTargetInfo;
+    pti->patchTC();
+  }
+
   new_ci->setTarget (pTargetInfo);
 
   new_ci->getPreprocessorOpts().UsePredefines = false;
@@ -360,20 +604,33 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
     for (clang::SourceManager::fileinfo_iterator fi = SM.fileinfo_begin(),
                                                  fe = SM.fileinfo_end();
          fi != fe; ++fi) {
-      const clang::FileEntry *file_entry = fi->first;
+      auto file_entry = fi->first;
       if (_closed_files.find (ACFileID (file_entry)) == _closed_files.end ()) {
         if (SM.isFileOverridden(file_entry) &&
             !new_SM.isFileOverridden(file_entry)) {
           // save changes in 'file_entry->getName ()' in the new source manager
+
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+          auto opt_mb = SM.getMemoryBufferForFileOrNone (fi->first);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+          new_SM.overrideFileContents(fi->first,
+            llvm::MemoryBuffer::getMemBufferCopy(opt_mb.value().getBuffer(),
+            opt_mb.value().getBufferIdentifier()));
+#else
+          new_SM.overrideFileContents(fi->first,
+            llvm::MemoryBuffer::getMemBufferCopy(opt_mb.getValue().getBuffer(),
+            opt_mb.getValue().getBufferIdentifier()));
+#endif
+#else
           const llvm::MemoryBuffer *buf = fi->second->getRawBuffer();
-          std::unique_ptr<llvm::MemoryBuffer> bcopy (llvm::MemoryBuffer::getMemBufferCopy(
-              buf->getBuffer(), buf->getBufferIdentifier()));
-          new_SM.overrideFileContents(fi->first, std::move(bcopy));
+          new_SM.overrideFileContents(fi->first,
+            llvm::MemoryBuffer::getMemBufferCopy(buf->getBuffer(), buf->getBufferIdentifier()));
+#endif
         }
       }
     }
     _closed_files.clear ();
-    string main_file = SM.getFileEntryForID (SM.getMainFileID ())->getName ().str ();
+    string main_file = ACFileID::main_file(SM).name();
     new_ci->InitializeSourceManager(clang::FrontendInputFile(main_file, clang::InputKind(clang::Language::CXX)));
     _ci->setFileManager (0);
   }
@@ -386,28 +643,28 @@ void ACProject::create_compiler_instance (ACConfig &conf) {
 
 // Add a forced include to the front end
 void ACProject::add_forced_include (const string &file) {
-  Filename incname;
-  if (!SysCall::canonical (Filename(file.c_str()), incname))
+  std::string incname;
+  if (!SysCall::make_canonical_path(file, incname))
     return; // error, shouldn't happen
 #ifdef FRONTEND_PUMA
-  config ().Add ("--include", incname.name ());
+  config ().Add ("--include", incname.c_str());
 #else
   std::vector<std::string> &incs = _ci->getInvocation ().getPreprocessorOpts ().Includes;
-  incs.push_back (incname.name ());
+  incs.push_back (incname);
 #endif
 }
 
 //Remove a forced include from the front end
 void ACProject::remove_forced_include (const string &file) {
-  Filename incname;
-  if (!SysCall::canonical (Filename(file.c_str()), incname))
+  std::string incname;
+  if (!SysCall::make_canonical_path(file, incname))
     return; // error, shouldn't happen
 #ifdef FRONTEND_PUMA
   for (unsigned int i = 0; i < config ().Options (); i++) {
     const ConfOption *o = config().Option (i);
     if (string (o->Name ()) != "--include" || o->Arguments () != 1)
       continue;
-    if (string(incname.name()) == o->Argument (0)) {
+    if (incname == o->Argument (0)) {
       config().Remove (o);
       break;
     }
@@ -415,7 +672,7 @@ void ACProject::remove_forced_include (const string &file) {
 #else
   vector<string> &incs = _ci->getInvocation ().getPreprocessorOpts ().Includes;
   for (vector<string>::iterator i = incs.begin (); i != incs.end (); ++i) {
-    if (string(incname.name()) == *i) {
+    if (incname == *i) {
       incs.erase (i);
       break;
     }
@@ -459,10 +716,11 @@ void ACProject::get_forced_includes (vector<string> &forced_includes) {
 
 Puma::Filename ACProject::getRelInclString (Puma::Filename from, Puma::Filename to) {
 
-  // start be calculating absolute canonical paths for 'to'
-  Filename to_abs;
-  if (!SysCall::canonical (to, to_abs))
+  // start by calculating the absolute canonical path for 'to'
+  std::string to_abs_str;
+  if (!SysCall::make_canonical_path(to.name(), to_abs_str))
     return to;
+  Puma::Filename to_abs(to_abs_str.c_str());
 
 //  cout << "to: " << to_abs.name() << endl;
 //  cout << "from: " << from_abs.name() << endl;
@@ -470,13 +728,14 @@ Puma::Filename ACProject::getRelInclString (Puma::Filename from, Puma::Filename
   // as a first strategy try to find an include path relative to a
   // search path (-I option)
   for (unsigned i = config ().Options (); i > 0; i--) {
-    const ConfOption *o = config ().Option (i-1);
+    const Puma::ConfOption *o = config ().Option (i-1);
     if (! strcmp (o->Name (), "-I")) {
       if (! o->Arguments ())
         continue;
-      Filename canon_inc_dir;
-      if (!SysCall::canonical (o->Argument (0), canon_inc_dir))
+      std::string canon_inc_dir_str;
+      if (!SysCall::make_canonical_path(o->Argument (0), canon_inc_dir_str))
         continue;
+      Puma::Filename canon_inc_dir(canon_inc_dir_str.c_str());
 
       int len = strlen (canon_inc_dir.name ());
       if (strncmp (canon_inc_dir.name (), to_abs.name (), len) == 0)
@@ -486,9 +745,10 @@ Puma::Filename ACProject::getRelInclString (Puma::Filename from, Puma::Filename
   }
 
   // alternatively generate a path from 'from' to 'to'.
-  Filename from_abs;
-  if (!SysCall::canonical (from, from_abs))
+  std::string from_abs_str;
+  if (!SysCall::make_canonical_path(from.name(), from_abs_str))
     return to_abs;
+  Puma::Filename from_abs(from_abs_str.c_str());
 
   // first find the point where the two absolute paths differ, e.g.
   // "/foo/bar/abcdef/file.h"
@@ -519,149 +779,29 @@ Puma::Filename ACProject::getRelInclString (Puma::Filename from, Puma::Filename
 }
 
 #ifdef FRONTEND_CLANG
-bool ACProject::isNewer (const char *file) const {
-
-  FileInfo fileinfo;
-  if (! SysCall::stat (file, fileinfo))
-    return false; // File does not exists.
-  long last_modified = fileinfo.modi_time ();
-
-  // determine the destination path of the file
-  ProjectFile::MapConstIter iter;
-  if (!PathManager::isBelow (file, iter)) {
-    assert (false); // if we came here, the file should be registered
-    return false;
-  }
-
-  ProjectFile &project_file = (ProjectFile&)iter->second;
-  Filename dest = project_file.dest ();
-  if (!dest.is_defined ()) {
-    // determine the destination path
-    std::ostringstream path;
-    if (!getDestinationPath (file, path))
-      return false;
-    string dest_path = path.str ();
-    project_file.dest (dest_path.c_str ());
-    dest = project_file.dest ();
-  }
-
-  bool newer = true;
-  if (SysCall::stat (dest.name (), fileinfo))
-    if (last_modified <= fileinfo.modi_time ())
-      newer = false;
-
-  return newer;
-}
 
 // Save all files that have been opened
 void ACProject::save () const {
   clang::SourceManager &SM = _ci->getSourceManager();
+
   for (clang::SourceManager::fileinfo_iterator fi = SM.fileinfo_begin(),
                                                fe = SM.fileinfo_end();
        fi != fe; ++fi) {
-    const clang::FileEntry *file_entry = fi->first;
-    if (_closed_files.find(ACFileID(file_entry)) == _closed_files.end())
-      save (file_entry, SM.isFileOverridden (file_entry));
-  }
-}
-
-bool ACProject::is_protected (clang::FileEntry *file_entry) {
-  // TODO: implement this function
-  return false;
-}
-
-
-bool ACProject::make_dir_hierarchy (Filename path) const {
-
-  // Remember where we are.
-  char cwd[PATH_MAX];
-  if (SysCall::getcwd (cwd, PATH_MAX, &err ()) == 0)
-    return false;
-
-  // Change into the directory and create every missing one
-  char *path_copy = StrCol::dup (path.name ());
-  char *p = path_copy;
-
-  // TODO: here is a problem with different drives on Windows!!!
-  if (path.is_absolute ()) {
-    SysCall::chdir (path.root (), &err ());
-    p += strlen (path.root ());
-  }
-
-  char *curr = p;
-  while (*p != '\0') {
-    if (*p == '/') {
-      *p = '\0';
-      if (strlen (curr) && ! make_dir (curr)) {
-        delete[] path_copy;
-        return false;
-      }
-      curr = p + 1;
+    auto file_entry = fi->first;
+    ACFileID file_id(file_entry);
+    if (_closed_files.find(file_id) == _closed_files.end()) {
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+      auto mb = SM.getMemoryBufferForFileOrNone(file_entry);
+#else
+      const llvm::MemoryBuffer *mb = SM.getMemoryBufferForFile(file_entry);
+#endif
+      ProjectFileManager::save(file_id.name().c_str(), SM.isFileOverridden (file_entry), mb->getBufferStart ());
     }
-    p++;
   }
-
-  delete[] path_copy;
-
-  // Go back
-  SysCall::chdir (cwd, &err ());
-
-  return true;
 }
 
-bool ACProject::make_dir (const char *directory) const {
-  while (!SysCall::chdir (directory))
-    if (! SysCall::mkdir (directory, &err ())) {
-      err () << sev_error << "unable to create directory \""
-             << directory << "\"." << endMessage;
-      return false;
-    }
-  return true;
-}
-
-// save an opened file to its destination
-void ACProject::save (const clang::FileEntry *file_entry, bool is_modified) const {
-  clang::SourceManager &SM = _ci->getSourceManager();
-  const char *name = file_entry->getName ().data();
-  assert (name);
-
-  // Do not write files to protected paths or files from outside the
-  // source directories.
-  if (isProtected (name))
-    return;
-
-  // determine the destination path of the file
-  ProjectFile::MapConstIter iter;
-  if (!PathManager::isBelow (name, iter)) {
-    assert (false); // if we came here, the file should be registered
-    return;
-  }
-
-  ProjectFile &project_file = (ProjectFile&)iter->second;
-  Filename dest = project_file.dest ();
-  if (!dest.is_defined ()) {
-    // determine the destination path
-    std::ostringstream path;
-    if (!getDestinationPath (name, path))
-      return;
-    string dest_path = path.str ();
-    project_file.dest (dest_path.c_str ());
-    dest = project_file.dest ();
-  }
-
-  // make sure that the directory for the file exists
-  if (!make_dir_hierarchy (dest))
-    return;
-
-  // Check whether the file has to be updated.
-  if (isNewer(name) || is_modified) {
-    // Write the file to disk.
-    char *file = (char*)dest.name ();
-    ofstream out (file, ios::out);
-    const llvm::MemoryBuffer *mb = SM.getMemoryBufferForFile (file_entry);
-    out << mb->getBufferStart ();
-  }
-
+void ACProject::close (ACFileID fid) {
+  _closed_files.insert (fid);
 }
 
 #endif
