@@ -30,8 +30,8 @@ bool ClangASTConsumer::isInProject(clang::Decl *D) {
   PresumedLoc PL = _sm->getPresumedLoc(D->getLocation());
   StringRef Name = PL.getFilename();
   StringRef BufferName = _sm->getBufferName(D->getLocation());
-  assert(!Name.startswith("<intro") || BufferName.startswith("<intro"));
-  return BufferName.startswith("<intro") ||
+  assert(!CLANG_STR_STARTS_WITH(Name, "<intro") || CLANG_STR_STARTS_WITH(BufferName,"<intro"));
+  return CLANG_STR_STARTS_WITH(BufferName, "<intro") ||
          (!Name.empty() && _model.get_project().isBelow(Name.str().c_str()));
 }
 
@@ -184,7 +184,7 @@ bool ClangASTConsumer::VisitVarDecl( VarDecl *VD ) {
 
   // if this is an injected variable that represents an argument of an
   // annotation, handle it accordingly
-  if (VD && VD->getName().startswith("__ac_attr_")) {
+  if (VD && CLANG_STR_STARTS_WITH(VD->getName(), "__ac_attr_")) {
     _model.register_annotation_param(VD);
     return true;
   }
@@ -382,12 +382,15 @@ bool ClangASTConsumer::TraverseStmt( clang::Stmt* S ) {
     AS = llvm::dyn_cast<clang::AttributedStmt>(S);
   if (_context_stack.isCurrentContextValid())
     FD = llvm::dyn_cast<clang::FunctionDecl>( _context_stack.getCurrentDecl() );
+
   bool handle_attr_stmt = false;
   if (AS && FD && isInProject(FD)) {
-    handle_attr_stmt = true;
-    ACM_Statement *stmt =
-        _model.register_statement(AS, _stmt_stack.empty() ? _parent_map[FD] : _stmt_stack.back());
-    _stmt_stack.push_back (stmt);
+    ACM_Name *parent = _stmt_stack.empty() ? _parent_map[FD] : _stmt_stack.back();
+    if (parent) {
+      handle_attr_stmt = true;
+      ACM_Statement *stmt = _model.register_statement(AS, parent);
+      _stmt_stack.push_back (stmt);
+    }
   }
 
   bool result = RecursiveASTVisitor<ClangASTConsumer>::TraverseStmt( S );
