@@ -867,7 +867,7 @@ void ClangFlowAnalysis::do_target_object_analysis(clang::DeclaratorDecl* func) {
     TransferFunctions TF(vals_begin, _target_obj_lids, _analysis_decl_ctx->getASTContext());
     
     for (CFGBlock::const_iterator BI=block->begin(), BE=block->end(); BI != BE; ++BI) {
-      if (Optional<CFGStmt> stmt = BI->getAs<CFGStmt>()) {
+      if (auto stmt = BI->getAs<CFGStmt>()) {
         TF.TraverseStmt(const_cast<Stmt*>(stmt->getStmt()));
       }
     }
