@@ -45,7 +45,11 @@ class InclusionRewriter : public PPCallbacks {
   Preprocessor &PP; ///< Used to find inclusion directives.
   SourceManager &SM; ///< Used to read and manage source files.
   raw_ostream &OS; ///< The destination stream for rewritten contents.
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  llvm::MemoryBufferRef PredefinesBuffer; ///< The preprocessor predefines.
+#else
   const llvm::MemoryBuffer *PredefinesBuffer; ///< The preprocessor predefines.
+#endif
   bool ShowLineMarkers; ///< Show #line markers.
   bool UseLineDirective; ///< Use of line directives or line markers.
   typedef std::map<unsigned, FileChange> FileChangeMap;
@@ -57,38 +61,61 @@ public:
   InclusionRewriter(Preprocessor &PP, raw_ostream &OS, bool ShowLineMarkers,
                     FileExpansionStrategy *strategy = 0);
   bool Process(FileID FileId, SrcMgr::CharacteristicKind FileType);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  void setPredefinesBuffer(const llvm::MemoryBufferRef &Buf) {
+#else
   void setPredefinesBuffer(const llvm::MemoryBuffer *Buf) {
+#endif
     PredefinesBuffer = Buf;
   }
 private:
   virtual void FileChanged(SourceLocation Loc, FileChangeReason Reason,
                            SrcMgr::CharacteristicKind FileType,
-                           FileID PrevFID);
-  virtual void FileSkipped(const FileEntry &ParentFile,
+                           FileID PrevFID) override;
+  virtual void FileSkipped(const FileEntryRef &ParentFile,
                            const Token &FilenameTok,
-                           SrcMgr::CharacteristicKind FileType);
+                           SrcMgr::CharacteristicKind FileType) override;
+
+
   virtual void InclusionDirective(SourceLocation HashLoc,
                                   const Token &IncludeTok,
                                   StringRef FileName,
                                   bool IsAngled,
                                   CharSourceRange FilenameRange,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+                                  clang::OptionalFileEntryRef File,
+#else
                                   const FileEntry *File,
+#endif
                                   StringRef SearchPath,
                                   StringRef RelativePath,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_19_1_6
+                                  const Module *SuggestedModule,
+                                  bool ModuleImported,
+#else
                                   const Module *Imported,
+#endif
                                   SrcMgr::CharacteristicKind FileType
-				                         );
+				                         ) override;
   void WriteLineInfo(const char *Filename, int Line,
                      SrcMgr::CharacteristicKind FileType,
                      StringRef EOL, StringRef Extra = StringRef());
   void WriteImplicitModuleImport(const Module *Mod, StringRef EOL);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  void OutputContentUpTo(const MemoryBufferRef &FromFile,
+#else
   void OutputContentUpTo(const MemoryBuffer &FromFile,
+#endif
                          unsigned &WriteFrom, unsigned WriteTo,
                          StringRef EOL, int &lines,
                          bool EnsureNewline = false,
                          bool CommentOut = false);
   void CommentOutDirective(Lexer &DirectivesLex, const Token &StartToken,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+                           const MemoryBufferRef &FromFile, StringRef EOL,
+#else
                            const MemoryBuffer &FromFile, StringRef EOL,
+#endif
                            unsigned &NextToWrite, int &Lines);
   bool HandleHasInclude(FileID FileId, Lexer &RawLex,
                         const DirectoryLookup *Lookup, Token &Tok,
@@ -105,7 +132,10 @@ private:
 InclusionRewriter::InclusionRewriter(Preprocessor &PP, raw_ostream &OS,
                                      bool ShowLineMarkers,
                                      FileExpansionStrategy *strategy)
-    : PP(PP), SM(PP.getSourceManager()), OS(OS), PredefinesBuffer(0),
+    : PP(PP), SM(PP.getSourceManager()), OS(OS),
+#if CLANG_VERSION_NUMBER < VERSION_NUMBER_12_0_0
+    PredefinesBuffer(0),
+#endif
     ShowLineMarkers(ShowLineMarkers),
     LastInsertedFileChange(FileChanges.end()),
     aspect(strategy) {
@@ -170,7 +200,7 @@ void InclusionRewriter::FileChanged(SourceLocation Loc,
 
 /// Called whenever an inclusion is skipped due to canonical header protection
 /// macros.
-void InclusionRewriter::FileSkipped(const FileEntry &/*ParentFile*/,
+void InclusionRewriter::FileSkipped(const FileEntryRef &/*ParentFile*/,
                                     const Token &FilenameTok,
                                     SrcMgr::CharacteristicKind /*FileType*/) {
   assert(LastInsertedFileChange != FileChanges.end() && "A file, that wasn't "
@@ -189,18 +219,33 @@ void InclusionRewriter::InclusionDirective(SourceLocation HashLoc,
                                            StringRef /*FileName*/,
                                            bool /*IsAngled*/,
                                            CharSourceRange /*FilenameRange*/,
-                                           const FileEntry * /*File*/,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+                                          clang::OptionalFileEntryRef /*File*/,
+#else
+                                          const FileEntry * /*File*/,
+#endif
                                            StringRef /*SearchPath*/,
                                            StringRef /*RelativePath*/,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_19_1_6
+                                           const Module *SuggestedModule,
+                                           bool ModuleImported,
+#else
                                            const Module *Imported,
+#endif
                                            SrcMgr::CharacteristicKind FileType
 					                                ) {
   assert(LastInsertedFileChange == FileChanges.end() && "Another inclusion "
     "directive was found before the previous one was processed");
   std::pair<FileChangeMap::iterator, bool> p = FileChanges.insert(
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_19_1_6
+    std::make_pair(HashLoc.getRawEncoding(), FileChange(HashLoc, SuggestedModule)));
+  assert(p.second && "Unexpected revisitation of the same include directive");
+  if (!ModuleImported)
+#else
     std::make_pair(HashLoc.getRawEncoding(), FileChange(HashLoc, Imported)));
   assert(p.second && "Unexpected revisitation of the same include directive");
   if (!Imported)
+#endif
     LastInsertedFileChange = p.first;
 }
 
@@ -216,7 +261,11 @@ InclusionRewriter::FindFileChangeLocation(SourceLocation Loc) const {
 
 /// Detect the likely line ending style of \p FromFile by examining the first
 /// newline found within it.
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+static StringRef DetectEOL(const MemoryBufferRef &FromFile) {
+#else
 static StringRef DetectEOL(const MemoryBuffer &FromFile) {
+#endif
   // detect what line endings the file uses, so that added content does not mix
   // the style
   const char *Pos = strchr(FromFile.getBufferStart(), '\n');
@@ -231,14 +280,22 @@ static StringRef DetectEOL(const MemoryBuffer &FromFile) {
 
 /// Writes out bytes from \p FromFile, starting at \p NextToWrite and ending at
 /// \p WriteTo - 1.
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+void InclusionRewriter::OutputContentUpTo(const MemoryBufferRef &FromFile,
+#else
 void InclusionRewriter::OutputContentUpTo(const MemoryBuffer &FromFile,
+#endif
                                           unsigned &WriteFrom, unsigned WriteTo,
                                           StringRef EOL, int &Line,
                                           bool EnsureNewline,
                                           bool CommentOut) {
   if (WriteTo <= WriteFrom)
     return;
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  if (FromFile == PredefinesBuffer) {
+#else
   if (&FromFile == PredefinesBuffer) {
+#endif
     // Ignore the #defines of the predefines buffer.
     WriteFrom = WriteTo;
     return;
@@ -278,7 +335,11 @@ void InclusionRewriter::OutputContentUpTo(const MemoryBuffer &FromFile,
 /// through the \p FromFile buffer.
 void InclusionRewriter::CommentOutDirective(Lexer &DirectiveLex,
                                             const Token &StartToken,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+                                            const MemoryBufferRef &FromFile,
+#else
                                             const MemoryBuffer &FromFile,
+#endif
                                             StringRef EOL,
                                             unsigned &NextToWrite, int &Line) {
   OutputContentUpTo(FromFile, NextToWrite,
@@ -366,21 +427,34 @@ bool InclusionRewriter::HandleHasInclude(
   // Now ask HeaderInfo if it knows about the header.
   // FIXME: Subframeworks aren't handled here. Do we care?
   bool isAngled = PP.GetIncludeFilenameSpelling(Tok.getLocation(), Filename);
-  const DirectoryLookup *CurDir;
-  const FileEntry *FileEnt = PP.getSourceManager().getFileEntryForID(FileId);
 
 // Create "Includers" for the following call of LookupFile depending on Clang version:
-  SmallVector<std::pair<const FileEntry *, const DirectoryEntry *>, 1>
-      Includers;
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+  clang::ConstSearchDirIterator *CurDir = nullptr;
+  auto FileEnt = PP.getSourceManager().getFileEntryRefForID(FileId);
+  std::pair< OptionalFileEntryRef, DirectoryEntryRef > Includers[] = { std::make_pair(FileEnt, FileEnt->getDir()) };
+#else
+  const DirectoryLookup *CurDir;
+  const FileEntry *FileEnt = PP.getSourceManager().getFileEntryForID(FileId);
+  SmallVector<std::pair<const FileEntry *, const DirectoryEntry *>, 1> Includers;
   Includers.push_back(std::make_pair(FileEnt, FileEnt->getDir()));
+#endif
 
   // Call LookupFile:
   // FIXME: Why don't we call PP.LookupFile here?
   auto File = PP.getHeaderSearchInfo().LookupFile(
       // StringRef Filename, SourceLocation IncludeLoc, bool isAngled,
       Filename, SourceLocation(), isAngled,
-      // const DirectoryLookup *FromDir, const DirectoryLookup *&CurDir,
-      nullptr, CurDir,
+      // const DirectoryLookup *FromDir,
+      nullptr,
+      // const DirectoryLookup *&CurDir,
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+      CurDir,
+#elif CLANG_VERSION_NUMBER >= VERSION_NUMBER_14_0_0
+      &CurDir,
+#else
+      CurDir,
+#endif
       // ArrayRef<std::pair<const FileEntry *, const DirectoryEntry *>> Includers,
       //    *or* (depending on Clang version (see above))
       // ArrayRef<const FileEntry *> Includers,
@@ -396,7 +470,11 @@ bool InclusionRewriter::HandleHasInclude(
       // , bool *IsFrameworkFound
       , nullptr
   );
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+  FileExists = File.has_value();
+#else
   FileExists = File.hasValue();
+#endif
   return true;
 }
 
@@ -406,11 +484,26 @@ bool InclusionRewriter::Process(FileID FileId,
                                 SrcMgr::CharacteristicKind FileType)
 {
   bool Invalid;
-  const MemoryBuffer &FromFile = *SM.getBuffer(FileId, &Invalid);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  auto opt_mb = SM.getBufferOrNone (FileId);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+  Invalid = !opt_mb.has_value();
+  llvm::MemoryBufferRef FromFile = opt_mb.value();
+#else
+  Invalid = !opt_mb.hasValue();
+  llvm::MemoryBufferRef FromFile = opt_mb.getValue();
+#endif
+#else
+  const llvm::MemoryBuffer &FromFile = *SM.getBuffer(FileId, &Invalid);
+#endif
   if (Invalid) // invalid inclusion
     return false;
   const char *FileName = FromFile.getBufferIdentifier().data();
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  Lexer RawLex(FileId, FromFile, PP.getSourceManager(), PP.getLangOpts());
+#else
   Lexer RawLex(FileId, &FromFile, PP.getSourceManager(), PP.getLangOpts());
+#endif
   RawLex.SetCommentRetentionState(false);
 
   StringRef EOL = DetectEOL(FromFile);
@@ -427,7 +520,7 @@ bool InclusionRewriter::Process(FileID FileId,
 
   // Ignore UTF-8 BOM, otherwise it'd end up somewhere else than the start
   // of the resulting file.
-  if (FromFile.getBuffer().startswith("\xEF\xBB\xBF"))
+  if (CLANG_STR_STARTS_WITH(FromFile.getBuffer(), "\xEF\xBB\xBF"))
     NextToWrite = 3;
 
   Token RawToken;
@@ -449,10 +542,10 @@ bool InclusionRewriter::Process(FileID FileId,
           case tok::pp_import: {
             const FileChange *Change = FindFileChangeLocation(
                 HashToken.getLocation());
-//            if (Change) {
-//              if (Change->Id.isInvalid())
-//                cout << "Change ID invalid!" << endl;
-//            }
+            //  if (Change) {
+            //    if (Change->Id.isInvalid())
+            //      cout << "Change ID invalid!" << endl;
+            //  }
             if (aspect && Change && Change->Id.isValid() &&
                 !aspect->should_expand_file(FileId, Change->Id))
               break;
@@ -526,9 +619,17 @@ bool InclusionRewriter::Process(FileID FileId,
                   // Rewrite __has_include_next(x)
                 } else if (RawToken.getIdentifierInfo()->isStr(
                                "__has_include_next")) {
+                  
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+                  // Can be done, because the parameter 'Lookup' is not used by 'HandleHasInclude'.
+                  // However, this is strange! Better replace this code by the latest version
+                  // of the rewriter
+                  const DirectoryLookup *Lookup = nullptr;
+#else
                   const DirectoryLookup *Lookup = PP.GetCurDirLookup();
                   if (Lookup)
                     ++Lookup;
+#endif
 
                   if (!HandleHasInclude(FileId, RawLex, Lookup, RawToken,
                                         HasFile))
@@ -621,7 +722,11 @@ void clang::RewriteIncludesInInput(Preprocessor &PP, raw_ostream *OS,
   do {
     PP.Lex(Tok);
   } while (Tok.isNot(tok::eof));
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+  Rewrite->setPredefinesBuffer(SM.getBufferOrFake(PP.getPredefinesFileID()));
+#else
   Rewrite->setPredefinesBuffer(SM.getBuffer(PP.getPredefinesFileID()));
+#endif
   Rewrite->Process(PP.getPredefinesFileID(), SrcMgr::C_User);
   Rewrite->Process(SM.getMainFileID(), SrcMgr::C_User);
   OS->flush();
