@@ -24,9 +24,8 @@
 #include "ACPreprocessor.h"
 #include "ACModel/Utils.h"
 #include "ClangAnnotation.h"
-
-// Puma library includes
-#include "Puma/SysCall.h"
+#include "SysCall.h"
+#include "ClangErrorStream.h"
 
 #include "clang/AST/Decl.h"
 #include "clang/AST/DeclCXX.h"
@@ -35,21 +34,26 @@
 #include <clang/AST/ExprCXX.h>
 #include "clang/Frontend/CompilerInstance.h"
 
-// start phase 1
-void ClangModelBuilder::setup_phase1 (const string &tunit_name, int tunit_len) {
-  _tunit_file = newTUnit();
-  _tunit_file->set_filename(model_filename (tunit_name.c_str()));
-  _tunit_file->set_len(tunit_len);
-  _tunit_file->set_time(modification_time (tunit_name));
-  get_files().insert(_tunit_file);
-  file_map().insert(
-      ClangModelBuilder::FileMapPair(model_filename(tunit_name.c_str()), _tunit_file));
-  set_root(0); // TODO: to be done by libacmodel
-
-  register_namespace1(0, "::");
-
+ClangModelBuilder::ClangModelBuilder (ACErrorStream &err, ACConfig &conf, ACProject &project) :
+  _err (err), _conf (conf), _project(project), _annotation_context (0),
+  _disable_source_loc_registration(false) {
+  for (auto i = 0; i < project.numPaths(); i++) {
+    string canonical_path;
+    if (!SysCall::make_canonical_path(project.src(i), canonical_path)) {
+      err << sev_error << "input path '" << project.src(i) << "' cannot be used" << endMessage;
+      continue;
+    }
+    ACM_ProjectPath *project_path = newProjectPath();
+    project_path->set_canonical(canonical_path);
+    get_project_paths().insert(project_path);
+  }
 }
 
+bool ClangModelBuilder::set_disable_source_loc_registration(bool new_val) {
+  bool old_val = _disable_source_loc_registration;
+  _disable_source_loc_registration = new_val;
+  return old_val;
+}
 
 void ClangModelBuilder::advice_infos (ACM_Aspect *jpl_aspect) {
   // collect the advice nodes
@@ -67,7 +71,7 @@ void ClangModelBuilder::advice_infos (ACM_Aspect *jpl_aspect) {
 
     while (DI != DE) {
       if (clang::NamedDecl *ND = llvm::dyn_cast<clang::NamedDecl>(*DI))
-        if (llvm::StringRef(ND->getNameAsString()).startswith("__a"))
+        if (CLANG_STR_STARTS_WITH(llvm::StringRef(ND->getNameAsString()), "__a"))
           break;
       ++DI;
     }
@@ -96,55 +100,27 @@ void ClangModelBuilder::advice_infos (ACM_Aspect *jpl_aspect) {
 
     // tree is the tree of the function definition
     TI_AdviceCode::of(*advice_code)->decl (FD);
+
+    // finally set the context information for the advice code, which is
+    // only known after parsing in phase 2
+    AdviceCodeContext context = ACC_NONE;
+    if (TI_AdviceCode::of(*advice_code)->needs_joinpoint_obj())
+      context = (AdviceCodeContext)(context | ACC_OBJ);
+    if (TI_AdviceCode::of(*advice_code)->needs_joinpoint_type())
+      context = (AdviceCodeContext)(context | ACC_TYPE);
+    if (TI_AdviceCode::of(*advice_code)->needs_context_vars())
+      context = (AdviceCodeContext)(context | ACC_VARS);
+    advice_code->set_context(context);
   }
 }
 
-static string absolutePath (const char *name) {
-  Puma::Filename result;
-  if (!Puma::SysCall::canonical (name, result))
-    return string();
-
-  return result.name ();
-}
-
-string ClangModelBuilder::model_filename (const char *name) {
-  string fname = absolutePath (name);
-  if (fname.empty ())
-    return fname;
-  ACProject &prj = _conf.project ();
-  // if the file does not belong to the project return the absolute path
-  if (!prj.isBelow (fname.c_str ()))
-    return fname;
-  // iterate over all project paths
-  for (long p = 0; p < prj.numPaths (); p++) {
-    Puma::Filename dir_abs;
-    if (!Puma::SysCall::canonical (prj.src (p), dir_abs)) {
-      assert (false);
-      return fname;
-    }
-    int dir_len = strlen (dir_abs.name ());
-    if (strncmp (dir_abs.name (), fname.c_str (), dir_len) == 0) {
-      return fname.substr(dir_len + 1);
-    }
-  }
-  // the file has to be below any of the directories => fatal error here
-  // assert (false); acgen.c does not belong to the project with this test
-  return fname;
-}
-
-
-// get the modification time of a file (UNIX Epoch value)
 long ClangModelBuilder::modification_time (const string &filename) {
-  string abs_name = absolutePath (filename.c_str ());
-  if (abs_name.empty ())
-    return -1; // File does not exists or other error
-  Puma::FileInfo fileinfo;
-  if (! Puma::SysCall::stat (abs_name.c_str (), fileinfo))
-    return -1; // File does not exists.
-  return fileinfo.modi_time ();
+  time_t result;
+  if (!SysCall::file_modification_time(filename, result))
+    return -1;
+  return result;
 }
 
-
 bool ClangModelBuilder::inside_template (clang::DeclContext *scope) const {
   if (llvm::isa<clang::TranslationUnitDecl>(scope))
     return false;
@@ -201,9 +177,9 @@ bool ClangModelBuilder::is_valid_model_class (clang::RecordDecl *ci) const {
   llvm::StringRef Name = ci->getName();
   std::string QualName = ci->getQualifiedNameAsString();
   if (Name.empty() ||
-      Name.startswith("%") || Name == "JoinPoint" ||
-      Name.startswith("__TJP") || Name == "__TI" ||
-      llvm::StringRef(QualName).startswith("AC::"))
+      CLANG_STR_STARTS_WITH(Name, "%") || Name == "JoinPoint" ||
+      CLANG_STR_STARTS_WITH(Name, "__TJP") || Name == "__TI" ||
+      CLANG_STR_STARTS_WITH(llvm::StringRef(QualName), "AC::"))
     return false;
 
   // Templates and classes nested in template class are not considered for
@@ -230,8 +206,8 @@ bool ClangModelBuilder::is_valid_model_function (clang::FunctionDecl *fi) const
 
   std::string S = fi->getNameAsString();
   llvm::StringRef Name = S;
-  if (Name.startswith("%a") || Name.startswith("__a") || Name == "aspectof" ||
-      Name == "aspectOf" || Name.startswith("__builtin_"))
+  if (CLANG_STR_STARTS_WITH(Name, "%a") || CLANG_STR_STARTS_WITH(Name, "__a") || Name == "aspectof" ||
+      Name == "aspectOf" || CLANG_STR_STARTS_WITH(Name, "__builtin_"))
     return false;
 
   // Explicitly filter the operators added by phase 1.
@@ -299,7 +275,7 @@ bool ClangModelBuilder::is_valid_model_namespace (clang::NamespaceDecl *ni) cons
   llvm::StringRef Name = ni->getName();
   // no template instance namespace, but anonymous namespaces!
   if ((ni->isAnonymousNamespace() || Name.find('<') == llvm::StringRef::npos) &&
-      Name != "AC" && !Name.startswith("__puma"))
+      Name != "AC" && !CLANG_STR_STARTS_WITH(Name, "__puma"))
     return true;
   return false;
 }
@@ -314,8 +290,13 @@ bool ClangModelBuilder::is_valid_model_type( const clang::QualType& qual_type )
   // these types shall not be used as a template-argument for a template type-parameter."
 
   return !qual_type->hasUnnamedOrLocalType()
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+      && qual_type->getLinkage() != clang::Linkage::None
+      && qual_type->getLinkage() != clang::Linkage::VisibleNone; // TODO: really?
+#else
       && qual_type->getLinkage() != clang::Linkage::NoLinkage
       && qual_type->getLinkage() != clang::Linkage::VisibleNoLinkage; // TODO: really?
+#endif
 
   // TODO: An enum type could be treated as a suitable integer type like the following:
   // Get the suitable integer type from the operator expression and the argument expressions with:
@@ -354,93 +335,6 @@ TU_Arg *ClangModelBuilder::register_arg (clang::QualType ti, llvm::StringRef nam
   return new_elem;
 }
 
-TU_Attribute *ClangModelBuilder::register_attrdecl1 (ACM_Namespace *parent, const std::string &name)
-{
-  TU_Attribute *elem = 0;
-  if (parent)
-    elem = (TU_Attribute*)map_lookup(*parent, name);
-  if (!elem || elem->type_val() != JPT_Attribute) {
-    elem = newAttribute();
-    elem->set_name(name);
-    if(parent->get_name() == "clang" || parent->get_name() == "gnu")
-      elem->set_builtin(true);
-    else
-      elem->set_builtin(false);
-    if (parent) {
-      map_insert(*parent, *elem, name);
-      parent->get_children().insert(elem);
-    }
-  }
-  else {
-    //for already registered attributes?
-    elem = 0;
-  }
-
-  return elem;
-}
-
-// find an attribute declaration
-TU_Attribute *ClangModelBuilder::find_attrdecl (const std::vector<std::string> &qual_name) {
-  ACM_Name *scope = get_root();
-  ACM_Name *found = nullptr;
-  for (string name : qual_name) {
-    found = map_lookup(*scope, name);
-    if (!found)
-      return nullptr;
-    scope = found;
-  }
-  if (found->type_val() != JPT_Attribute)
-    return nullptr;
-  return (TU_Attribute*)found;
-}
-
-ACM_Pointcut *ClangModelBuilder::register_pointcut1 (ACM_Name *parent, const string &name,
-    bool is_virtual, const string& expr) {
-
-  TU_Pointcut *elem = 0;
-  if (parent)
-    elem = (TU_Pointcut*)map_lookup(*parent, name);
-  if (!elem || elem->type_val() != JPT_Pointcut) {
-    elem = newPointcut();
-    elem->set_name(name);
-    elem->set_expr (expr);
-    elem->set_builtin(false);
-    elem->set_kind(PT_NORMAL);
-    if (parent) {
-      if (is_virtual && expr == "0")
-        elem->set_kind(PT_PURE_VIRTUAL);
-      else if (is_virtual || overrides_virtual_pointcut(parent, name))
-        elem->set_kind(PT_VIRTUAL);
-      map_insert(*parent, *elem, name);
-      parent->get_children().insert(elem);
-    }
-  }
-  else {
-    elem = 0;
-  }
-
-  return elem;
-}
-
-bool ClangModelBuilder::overrides_virtual_pointcut (ACM_Name *parent, const string &name) {
-  if (parent->type_val() != JPT_Class && parent->type_val() != JPT_Aspect)
-    return false;
-  ACM_Class *cls = (ACM_Class*)parent;
-  typedef ACM_Container<ACM_Class, false> BList;
-  const BList &blist = cls->get_bases();
-  for (BList::const_iterator i = blist.begin (); i != blist.end(); ++i) {
-    ACM_Class *base = (ACM_Class*)*i;
-    ACM_Name *elem = map_lookup(*base, name);
-    if (!elem || elem->type_val() != JPT_Pointcut)
-      continue;
-    ACM_Pointcut *base_pct = (ACM_Pointcut*)elem;
-    if (base_pct->get_kind() == PT_PURE_VIRTUAL ||
-        base_pct->get_kind() == PT_VIRTUAL ||
-        overrides_virtual_pointcut (base, name))
-      return true;
-  }
-  return false;
-}
 
 TU_Attribute *ClangModelBuilder::register_attrdecl (clang::FunctionDecl *fi,
     ACM_Name *parent) {
@@ -452,7 +346,7 @@ TU_Attribute *ClangModelBuilder::register_attrdecl (clang::FunctionDecl *fi,
     return 0;
 
   // check if it has the correct kind of name
-  if (!fi->getIdentifier() || !fi->getName().startswith("__ac_attr1_"))
+  if (!fi->getIdentifier() || !CLANG_STR_STARTS_WITH(fi->getName(), "__ac_attr1_"))
     return 0;
 
   // build the name of the function
@@ -471,6 +365,21 @@ TU_Attribute *ClangModelBuilder::register_attrdecl (clang::FunctionDecl *fi,
     }
     elem->decl (def);
     add_source_loc (elem, def);
+    if (signature(*parent) == "::")
+      elem->set_builtin(true); // root scope => must be built-in, e.g. [[noreturn]]
+    else if (signature(*parent).find("::") != string::npos)
+      elem->set_builtin(false); // nested namespace => only possible with user-defined attributed
+    else {
+      // non-nested namespace => check whether the attribute is known to clang
+      clang::Preprocessor &PP = _project.get_compiler_instance()->getPreprocessor();
+      clang::IdentifierInfo *ii_scope = PP.getIdentifierInfo(signature(*parent));
+      clang::IdentifierInfo *ii_name  = PP.getIdentifierInfo(name);
+      bool is_user_defined = 
+        (clang::AttributeCommonInfo::getParsedKind(ii_name, ii_scope, clang::AttributeCommonInfo::Syntax::AS_CXX11) ==
+        clang::AttributeCommonInfo::Kind::UnknownAttribute);
+      elem->set_builtin(!is_user_defined);
+    }
+    // cout << "phase 2 " << signature(*parent) << " " << signature(*elem) << " " << elem->get_builtin() << endl;
   }
 
   return elem;
@@ -498,8 +407,8 @@ void ClangModelBuilder::register_annotation_param (clang::VarDecl *VD) {
   else if (!arg->isCXX11ConstantExpr(VD->getASTContext(), &val)) { // param is not constant
     _annotation_context->param_values.push_back ("!C");
     _annotation_context->param_expressions.push_back ("!C");
-//    _err << Puma::sev_error << _annotation_context->tokBegin
-//         << "argument of attribute is not constant." << Puma::endMessage;
+//    _err << sev_error << _annotation_context->tokBegin
+//         << "argument of attribute is not constant." << endMessage;
   }
   else {
     // everything fine: store the value
@@ -564,19 +473,23 @@ TU_Function *ClangModelBuilder::register_function (clang::FunctionDecl *fi,
   TU_Function *elem = (TU_Function*)map_lookup(*parent, sig);
   bool hidden_function = (elem && elem->type_val() == JPT_Function &&
                           ((isStatic && !isStaticMethod) &&
-                           (!elem->has_static_in () || elem->get_static_in () != _tunit_file)));
+                           (!elem->has_static_in () || elem->get_static_in () != tunit_file())));
   if (!elem || hidden_function) {
     elem = newFunction();
     elem->set_name(TI_Function::name(fi));
     if (isStatic && !isStaticMethod)
-      elem->set_static_in (_tunit_file);
+      elem->set_static_in (tunit_file());
     map_insert(*parent, *elem, sig);
     parent->get_children().insert(elem);
     FunctionType ft = FT_NON_MEMBER;
     if (llvm::isa<clang::CXXConstructorDecl>(fi))
       ft = FT_CONSTRUCTOR;
     else if (llvm::isa<clang::CXXDestructorDecl>(fi)) {
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+      if (method->isVirtual() && method->isPureVirtual())
+#else
       if (method->isVirtual() && method->isPure())
+#endif
         ft = FT_PURE_VIRTUAL_DESTRUCTOR;
       else if (method->isVirtual())
         ft = FT_VIRTUAL_DESTRUCTOR;
@@ -586,7 +499,11 @@ TU_Function *ClangModelBuilder::register_function (clang::FunctionDecl *fi,
     else if (method) {
       if (isStaticMethod)
         ft = FT_STATIC_MEMBER;
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+      else if (method->isVirtual() && method->isPureVirtual())
+#else
       else if (method->isVirtual() && method->isPure())
+#endif
         ft = FT_PURE_VIRTUAL_MEMBER;
       else if (method->isVirtual())
         ft = FT_VIRTUAL_MEMBER;
@@ -626,9 +543,13 @@ TU_Function *ClangModelBuilder::register_function (clang::FunctionDecl *fi,
     llvm::StringRef Name = PL.getFilename();
     llvm::StringRef BufferName = sm.getBufferName(fi->getLocation());
     if (!elem->get_builtin() &&
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+        (!method || !(method->isVirtual() && method->isPureVirtual())) &&
+#else
         (!method || !(method->isVirtual() && method->isPure())) &&
+#endif
         !(inside_template_instance (fi)) &&
-        (BufferName.startswith("<intro") ||
+        (CLANG_STR_STARTS_WITH(BufferName, "<intro") ||
          (!Name.empty() && _project.isBelow(Name.str().c_str())))) {
       if (llvm::isa<clang::CXXConstructorDecl>(def))
         register_construction (elem);
@@ -652,8 +573,7 @@ TU_Function *ClangModelBuilder::register_function (clang::FunctionDecl *fi,
   }
 
   // update the list of source locations
-  add_source_loc (elem, fi,
-                  fi->isThisDeclarationADefinition () ? SLK_DEF : SLK_DECL);
+  add_source_loc (elem, fi, fi->isThisDeclarationADefinition () ? SLK_DEF : SLK_DECL);
 
   // check and register attributes of this function
   if (fi->hasAttrs())
@@ -752,36 +672,6 @@ ACM_Statement *ClangModelBuilder::register_statement ( clang::AttributedStmt *as
   return elem;
 }
 
-ACM_ClassSlice *ClangModelBuilder::register_class_slice (ACM_Name *scope, string name,
-    bool is_struct) {
-
-  ACM_Name *found = map_lookup(*scope, name);
-  if (found && found->type_val() == JPT_ClassSlice)
-    return (ACM_ClassSlice*)found;
-
-  ACM_ClassSlice *new_elem = (ACM_ClassSlice*)newClassSlice();
-  new_elem->set_name(name);
-  new_elem->set_builtin(false);
-  new_elem->set_is_struct(is_struct);
-  map_insert(*scope, *new_elem, name);
-  scope->get_children().insert(new_elem);
-  return new_elem;
-}
-
-
-ACM_Class * ClangModelBuilder::register_class1 (ACM_Name *scope, string name, bool in_project) {
-  ACM_Name *found = map_lookup(*scope, name);
-  if (found && found->type_val() == JPT_Class)
-    return (TU_Class*)found;
-  ACM_Class *new_elem = newClass();
-  new_elem->set_name(name);
-  new_elem->set_builtin(false);
-  if (!in_project) new_elem->get_tunits().insert (_tunit_file);
-  new_elem->set_intro_target(in_project);
-  scope->get_children().insert(new_elem);
-  map_insert(*scope, *new_elem, name);
-  return new_elem;
-}
 
 void ClangModelBuilder::register_constructor_and_destructor (clang::RecordDecl *ci,
     ACM_Class *elem) {
@@ -855,7 +745,7 @@ void ClangModelBuilder::register_constructor_and_destructor (clang::RecordDecl *
 
 
 TU_Class *ClangModelBuilder::register_class (clang::RecordDecl *ci,
-                                        ACM_Name *parent, bool set_source) {
+                                        ACM_Name *parent) {
 
   // only classes are relevant
   if (!is_valid_model_class (ci))
@@ -878,43 +768,30 @@ TU_Class *ClangModelBuilder::register_class (clang::RecordDecl *ci,
     map_insert(*parent, *elem, name_with_template_args);
     bool intro_target = is_intro_target (ci);
     elem->set_intro_target (intro_target);
-    if (!intro_target) elem->get_tunits().insert (_tunit_file);
+    if (!intro_target) elem->get_tunits().insert (tunit_file());
   }
   else
     elem = (TU_Class*)found;
   // set the class attributes
   if (ci->isThisDeclarationADefinition ())
     elem->decl (ci);
-  if (set_source && elem->type_val () != JPT_Aspect) {
+  if (elem->type_val () != JPT_Aspect) {
     // FIXME: Is this right?
     if (!inside_template_instance(ci))
       add_source_loc (elem, ci,
-                      ci->isThisDeclarationADefinition () ? SLK_DEF : SLK_DECL);
+                      // ci->isThisDeclarationADefinition () ? SLK_DEF : SLK_DECL);
+                      ci->getDefinition() == ci? SLK_DEF : SLK_DECL);
   }
 
   register_constructor_and_destructor (ci, elem);
 
-  // check and register attributes of this namespace declaration
+  // check and register attributes of this class declaration
   if (ci->hasAttrs())
     handle_attributes(elem, ci->getAttrs(), ci->isFirstDecl());
 
   return elem;
 }
 
-ACM_Aspect * ClangModelBuilder::register_aspect1 (ACM_Name *scope, string name, bool in_project) {
-  ACM_Name *found = map_lookup(*scope, name);
-  if (found && found->type_val() == JPT_Aspect)
-    return (TU_Aspect*)found;
-  ACM_Aspect *new_elem = newAspect();
-  new_elem->set_name(name);
-  new_elem->set_builtin(false);
-  if (!in_project) new_elem->get_tunits().insert (_tunit_file);
-  new_elem->set_intro_target(in_project);
-  scope->get_children().insert(new_elem);
-  map_insert(*scope, *new_elem, name);
-  return new_elem;
-}
-
 
 TU_Aspect *ClangModelBuilder::register_aspect (clang::RecordDecl *ci, ACM_Name *parent) {
   // only classes are relevant
@@ -936,13 +813,24 @@ TU_Aspect *ClangModelBuilder::register_aspect (clang::RecordDecl *ci, ACM_Name *
   elem->decl (ci);
 
   register_constructor_and_destructor (ci, elem);
+
+  // check and register attributes of this aspect declaration
+  if (ci->hasAttrs())
+    handle_attributes(elem, ci->getAttrs(),
+      ci->getPreviousDecl() && ci->getPreviousDecl()->isFirstDecl()); // parameter first_decl -> check for second decl instead (first is generated)
+
   return elem;
 }
 
 
 bool ClangModelBuilder::is_intro_target (clang::Decl *decl) const {
   clang::SourceManager &SM = _project.get_compiler_instance()->getSourceManager();
-  const llvm::MemoryBuffer *buf = SM.getBuffer(SM.getFileID(decl->getLocation()));
+      clang::FileID fid = SM.getFileID(decl->getLocation());
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+      llvm::MemoryBufferRef buf = SM.getBufferOrFake(fid);
+#else
+      const llvm::MemoryBuffer *buf = SM.getBuffer(fid);
+#endif
   if (IntroductionUnit::cast(buf))
     return true;
 
@@ -953,29 +841,6 @@ bool ClangModelBuilder::is_intro_target (clang::Decl *decl) const {
 }
 
 
-ACM_Namespace * ClangModelBuilder::register_namespace1 (ACM_Name *scope, string name, bool in_project) {
-  ACM_Namespace *result = 0;
-  assert(scope || name == "::");
-  ACM_Name *found = (scope ? map_lookup(*scope, name) : get_root());
-  if (!found || found->type_val() != JPT_Namespace) {
-    result = newNamespace();
-    result->set_name(name);
-    result->set_builtin(false);
-    if (!in_project) result->get_tunits().insert (_tunit_file);
-    if (scope) {
-      scope->get_children().insert(result);
-      map_insert(*scope, *result, name);
-    }
-    else {
-      set_root(result);
-    }
-  }
-  else
-    result = (TU_Namespace*)found;
-  return result;
-}
-
-
 TU_Namespace *ClangModelBuilder::register_namespace (clang::NamespaceDecl *n,
   ACM_Name *parent) {
     // not all Puma namespaces should be registered
@@ -1196,7 +1061,7 @@ TU_Builtin* ClangModelBuilder::register_builtin_operator_call(clang::Expr* built
 
     // Mark the function as 'used' by this translation unit,
     // to (hopefully) avoid garbage collection on xml-model merge
-    model_operator_function->get_tunits ().insert (_tunit_file);
+    model_operator_function->get_tunits ().insert (tunit_file());
   }
 
   // Create the operator-call:
@@ -1253,8 +1118,8 @@ ACM_Access *ClangModelBuilder::register_call (clang::FunctionDecl *called,
   else {
     lexical = register_variable (vd_caller);
     if (!lexical) {
-      _err << Puma::sev_warning << CLANG_GET_LOC_START(call_node)
-           << "location of function call invalid" << Puma::endMessage;
+      _err << sev_warning << CLANG_GET_LOC_START(call_node)
+           << "location of function call invalid" << endMessage;
       return 0;
     }
   }
@@ -1601,7 +1466,11 @@ void ClangModelBuilder::add_source_loc (ACM_Any *name, clang::Stmt *tree,
 void ClangModelBuilder::add_source_loc(ACM_Any *name, clang::SourceLocation location,
                                   clang::SourceLocation endlocation,
                                   SourceLocKind kind) {
+  // while we are in the introducer the length of source locations is usually invalid -> don't add it
+  if (_disable_source_loc_registration)
+    return;
   assert(name);
+
   clang::SourceManager &SM = _project.get_compiler_instance()->getSourceManager();
 
   // check if this file belong to our project
@@ -1626,65 +1495,57 @@ void ClangModelBuilder::add_source_loc(ACM_Any *name, clang::SourceLocation loca
       const Container &tunits = jpl_name->get_tunits ();
       bool found = false;
       for (Container::const_iterator i = tunits.begin(); i != tunits.end(); ++i) {
-        if (*i == _tunit_file) {
+        if (*i == tunit_file()) {
           found = true;
           break;
         }
       }
       if (!found)
-        jpl_name->get_tunits ().insert (_tunit_file);
+        jpl_name->get_tunits ().insert (tunit_file());
     }
     return;
   }
 
+  // register, if the location is in an unknown (header) file
   clang::FileID ID = SM.getFileID(location);
-  ClangModelBuilder::FileMap::iterator i =
-      file_map().find(model_filename(PL.getFilename()));
-  ACM_File *file = 0;
-  if (i != file_map ().end ())
-    file = i->second;
-  else {
-    // TODO: temporary hack
-    int len = SM.getSpellingLineNumber(SM.getLocForEndOfFile(ID));
-    // TODO: in the future, handle aspect headers differently
-    ACM_Header *new_file = newHeader();
-    get_files().insert(new_file);
-    new_file->set_filename(model_filename(PL.getFilename()));
-    new_file->set_len(len);
-    new_file->get_in().insert(_tunit_file);
-    if (const clang::FileEntry *fe = SM.getFileEntryForID(ID))
-      new_file->set_time(fe->getModificationTime());
-    else
-      new_file->set_time(0);
-    file = new_file;
-    file_map().insert(
-        ClangModelBuilder::FileMapPair(model_filename(PL.getFilename()), file));
-  }
+  int file_len = SM.getSpellingLineNumber(SM.getLocForEndOfFile(ID));
+  int modi_time = 0;
+  if (const clang::FileEntry *fe = SM.getFileEntryForID(ID))
+    modi_time = fe->getModificationTime();
+  ACM_File * file = register_file(PL.getFilename(), file_len, modi_time, true);
+
   int line = PL.getLine();
   int len = PLEnd.getLine() - line + 1;
   // TODO: really use linear search here?
   typedef ACM_Container<ACM_Source, true> Container;
   const Container &sources = name->get_source ();
   bool found = false;
+  // if (name->type_val() & JPT_Name) {
+  //   ACM_Name *n = (ACM_Name*)name;
+  //   cout << n->type_str() << " " << n->get_name() << ": " << kind << " " << line << " " << len << " " << file->get_filename() << endl;
+  // }
   for (Container::const_iterator i = sources.begin(); i != sources.end(); ++i) {
     if ((*i)->get_line () == line &&
         (*i)->get_file () == file &&
-        (*i)->get_kind () == kind &&
-        (*i)->get_len () == len) {
-//      if ((*i)->get_len () != len)
-//        _err << sev_warning << "Internal problem: Length differs for source at same location (file='" << file->get_filename().c_str () << "', line="
-//        << line << ")" << endMessage;
+        (*i)->get_kind () == kind) {
+      if ((*i)->get_len () != len) {
+        _err << sev_warning << "project model problem: length of source differs at same location (file='"
+        << file->get_filename().c_str () << "', line=" << line << ")" << endMessage;
+        // assert(false);
+      }
       found = true;
       break;
     }
   }
   if (!found) {
-    ACM_Source *source = newSource();
+    TU_Source *source = newSource();
     assert(file);
     source->set_file(file);
     source->set_line(line);
     source->set_len(len);
     source->set_kind(kind);
+    source->set_start_loc(clang::FullSourceLoc(location, SM));
+    source->set_end_loc(clang::FullSourceLoc(endlocation, SM));
     name->get_source().insert(source);
   }
 }
@@ -1699,7 +1560,8 @@ void ClangModelBuilder::handle_attributes (ACM_Name *name, const clang::AttrVec
 
   vector<string> attrStructNames;
   for(const clang::Attr *attr_use : attrs) {
-    AnnotationMap::iterator iter = annotation_map().find(attr_use->getLocation());
+    // this is weird: the end of the range is the location stored earlier in the map
+    AnnotationMap::iterator iter = annotation_map().find(attr_use->getRange().getEnd());
     if (iter == annotation_map().end())
       continue;
     Annotation &annotation = iter->second;
@@ -1717,16 +1579,16 @@ void ClangModelBuilder::handle_attributes (ACM_Name *name, const clang::AttrVec
     if (!attr) {
       // use of undeclared attributes is only allowed if it is not user-defined
       if (annotation.is_user_defined)
-        _err << Puma::sev_error << attr_use->getLocation() << "undeclared attribute '"
-          << annotation.get_qualified_name().c_str() << "' used" << Puma::endMessage;
+        _err << sev_error << attr_use->getLocation() << "undeclared attribute '"
+          << annotation.get_qualified_name().c_str() << "' used" << endMessage;
       continue;
     }
 
     // so far attribute parameters are not checked and ignored during the pointcut
     // evaluation => print a warning
     if (!annotation.params.empty())
-      _err << Puma::sev_warning << attr_use->getLocation() << "attribute parameters of '"
-        << annotation.get_qualified_name().c_str() << "' ignored" << Puma::endMessage;
+      _err << sev_warning << attr_use->getLocation() << "attribute parameters of '"
+        << annotation.get_qualified_name().c_str() << "' ignored" << endMessage;
 
     // check whether the annotation is already registered at this model element
     ACM_Annotation *found_anno = 0;
@@ -1740,9 +1602,9 @@ void ClangModelBuilder::handle_attributes (ACM_Name *name, const clang::AttrVec
       // in the first declaration of an entity all attributes must be present
       if (found_anno) {
         // the same attribute use more than once! => error
-        _err << Puma::sev_warning << attr_use->getLocation() << "attribute '"
+        _err << sev_warning << attr_use->getLocation() << "attribute '"
           << annotation.get_qualified_name().c_str()
-          << "' used more than once on the same entity" << Puma::endMessage;
+          << "' used more than once on the same entity" << endMessage;
       }
       else {
         ACM_Annotation *anno = newAnnotation();
@@ -1758,9 +1620,9 @@ void ClangModelBuilder::handle_attributes (ACM_Name *name, const clang::AttrVec
     }
     else {
       if (!found_anno) {
-        _err << Puma::sev_error << attr_use->getLocation() << "attribute '"
+        _err << sev_error << attr_use->getLocation() << "attribute '"
           << annotation.get_qualified_name().c_str()
-          << "' added after first declaration" << Puma::endMessage;
+          << "' added after first declaration" << endMessage;
       }
     }
   }
