@@ -97,8 +97,7 @@ const WeavePos &WeaverBase::weave_pos (clang::SourceLocation loc, int p) {
 }
 
 const WeavePos &WeaverBase::header_pos () {
-  clang::FileID main_fid = _rewriter->getSourceMgr().getMainFileID();
-  return header_pos (_rewriter->getSourceMgr().getFileEntryForID (main_fid));
+  return header_pos(ACFileID::main_file(_rewriter->getSourceMgr()));
 }
 
 const WeavePos &WeaverBase::header_pos (ACFileID acfid) {
@@ -110,8 +109,7 @@ const WeavePos &WeaverBase::header_pos (ACFileID acfid) {
 }
 
 const WeavePos &WeaverBase::footer_pos () {
-  clang::FileID main_fid = _rewriter->getSourceMgr().getMainFileID();
-  return footer_pos (_rewriter->getSourceMgr().getFileEntryForID (main_fid));
+  return footer_pos(ACFileID::main_file(_rewriter->getSourceMgr()));
 }
 
 const WeavePos &WeaverBase::footer_pos (ACFileID acfid) {
@@ -212,7 +210,7 @@ bool WeaverBase::lookup_macro_token (clang::SourceLocation loc, clang::SourceLoc
   clang::SourceLocation expansion_loc = SM.getExpansionLoc(loc);
   MacroMap::iterator mm_iter = _macros.find (expansion_loc);
   if (mm_iter == _macros.end ()) {
-    cout << "fatal: macro not found" << endl;
+    cout << "fatal: macro not found in expansion location map" << endl;
     return false;
   }
 
@@ -420,7 +418,12 @@ void WeaverBase::commit_internal () {
     if (nl) { // at least one new line character was inserted here
       // generate and insert a #line directive to make sure that debuggers will
       // be able to show the original source locations
-      const llvm::MemoryBuffer *u = SM.getBuffer(SM.getFileID(wp._loc));
+      clang::FileID fid = SM.getFileID(wp._loc);
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_12_0_0
+      llvm::MemoryBufferRef u = SM.getBufferOrFake(fid);
+#else
+      const llvm::MemoryBuffer *u = SM.getBuffer(fid);
+#endif
       if (IntroductionUnit::cast(u)) {
 	// TODO: Why don't we insert the #line directive into introduced code? Check!
       }
@@ -459,7 +462,7 @@ void WeaverBase::resolve_macros() {
       clang::CharSourceRange source_range = SM.getExpansionRange(somewhere_inside);
       std::pair<clang::SourceLocation, clang::SourceLocation> range =
           std::pair<clang::SourceLocation, clang::SourceLocation> (source_range.getBegin(), source_range.getEnd());
-    const clang::RewriteBuffer *RB = _rewriter->getRewriteBufferFor(fid);
+    const auto *RB = _rewriter->getRewriteBufferFor(fid);
     if (!RB) {
       cout << "fatal: Rewrite buffer not found" << endl;
       continue;
@@ -485,9 +488,15 @@ bool WeaverBase::commit (clang::SourceManager &Target) {
   for (clang::Rewriter::buffer_iterator i = _rewriter->buffer_begin(),
                                         e = _rewriter->buffer_end();
        i != e; ++i) {
+#if CLANG_VERSION_NUMBER >= VERSION_NUMBER_18_1_3
+    if (auto FE = SM.getFileEntryRefForID(i->first)) {
+      std::string data(i->second.begin(), i->second.end());
+      Target.overrideFileContents(FE.value(), llvm::MemoryBuffer::getMemBufferCopy(data, FE.value().getName()));
+#else
     if (const clang::FileEntry *FE = SM.getFileEntryForID(i->first)) {
       std::string data(i->second.begin(), i->second.end());
       Target.overrideFileContents(FE, llvm::MemoryBuffer::getMemBufferCopy(data, FE->getName()));
+#endif
     }
   }
 
@@ -510,7 +519,7 @@ static void po_intros(WeaverBase &wb, ::ACIntroducer &intro,
   if (!unit)
     return;
 
-  if (const clang::RewriteBuffer *RB =
+  if (const auto *RB =
           wb.getRewriter().getRewriteBufferFor(unit->file_id())) {
     std::string str(RB->begin(), RB->end());
     if (depth == 0) // Insert the last unit via weave_pos, there may be macros.
