@@ -1,6 +1,8 @@
 // Testing passing of "this" as argument of the call-wrapper if
 // "that" pointcut functions with runtime checks are used.
-// In this test a TJP-Struct should not be inserted.
+// Additionally, this test checks if the TJP-Struct incl. proceed
+// function is correctly inserted even if it is only used by the condition
+// check code.
 
 #include <iostream>
 using namespace std;
@@ -51,6 +53,15 @@ aspect Asp {
     cout << "after" << call_text << endl;
   }
   
+  advice call_func_that_B() : around() {
+    cout << "around 1" << call_text << endl;
+  }
+  
+  advice call_func_that_B() : around() {
+    cout << "around 2" << call_text << endl;
+  }
+  
+  
   advice execution_memberfunc_that_B() : before() {
     cout << "before" << execution_text << endl;
   }
@@ -58,4 +69,13 @@ aspect Asp {
   advice execution_memberfunc_that_B() : after() {
     cout << "after" << execution_text << endl;
   }
+  
+   advice execution_memberfunc_that_B() : around() {
+    cout << "around 1" << execution_text << endl;
+  }
+  
+  advice execution_memberfunc_that_B() : around() {
+    cout << "around 2" << execution_text << endl;
+  }
 };
+
